/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* 
 * Note to regenerate the Makefile and configure script:
 * aclocal
 * autoconf
 * touch NEWS README AUTHORS ChangeLog
 * automake -a
 */

#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "dbacl.h"

/* global variables */

long int default_max_token_bits = 16;
long int default_max_tokens = (1<<16);

Dirichlet dirichlet;

Learner learner;
Category cat[MAX_CAT];
int cat_count = 0;
int filter[MAX_CAT];
int filter_count = 0;
Empirical empirical;

char *textbuf = NULL;
int textbuf_len = 0;

/* for option processing */
extern char *optarg;
extern int optind, opterr, optopt;

unsigned long int options = 0;
char *title = "";

/* string hash function */
extern unsigned long int hash( unsigned char *k, unsigned long int length, unsigned long int initval);

int exit_code = 0; /* default */

/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

void usage(char **argv) {
    fprintf(stderr, "usage1: %s [-vnNR] [-T type] -c CATEGORY [-c CATEGORY]...\n", argv[0]);
    fprintf(stderr, "        [-f KEEP]... [FILE]...\n");
    fprintf(stderr, "       classifies FILE or STDIN using CATEGORY, optionally\n");
    fprintf(stderr, "       removing all lines which don't fit the category KEEP\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "usage2: %s [-vnrND] [-h size] [-T type] -l CATEGORY [FILE]...\n", argv[0]);
    fprintf(stderr, "       builds a maximum entropy model from the words in FILE or STDIN\n");
    fprintf(stderr, "usage3: %s -V\n", argv[0]);
    fprintf(stderr, "       prints program version\n");
}


/***********************************************************
 * EMPIRICAL DISTRIBUTION OF TOKENS                        *
 ***********************************************************/

/* initialize global learner object */
void init_empirical() {

    /* some constants */
    empirical.max_tokens = default_max_tokens;
    empirical.max_token_bits = default_max_token_bits;
    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.cached = 0;
    empirical.stack_top = 0;

    /* allocate room for hash */
    empirical.hash = calloc(empirical.max_tokens, sizeof(h_item));
    if( !empirical.hash ) {
	fprintf(stderr, "error: not enough memory? I couldn't allocate %li bytes\n", (sizeof(h_item) * empirical.max_tokens));
	exit(0);
    }
}

void clear_empirical() {
    int i;

    if( empirical.cached ) {
	/* this may actually be slower than a global memset */ 
	for(i = 0; i < empirical.stack_top; i++) {
	    memset(empirical.stack[i], 0, sizeof(h_item));
	}
    } else {
	memset(empirical.hash, 0, sizeof(h_item) * empirical.max_tokens);
    }

    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.stack_top = 0;
}

h_item *find_in_empirical(unsigned long int id) {
    h_item *i, *loop;
    /* start at id */
    i = loop = &empirical.hash[id & (empirical.max_tokens - 1)];

    while( FILLEDP(i) ) {
	if( EQUALP(i->id,id) ) {
	    return i; /* found id */
	} else {
	    i++; /* not found */
	    i = (i >= &empirical.hash[empirical.max_tokens]) ? empirical.hash : i; /* wrap around */
	    if( i == loop ) {
		return NULL; /* when hash table is full */
	    }
	}
    }

    /* empty slot, so not found */

    return i; 
}

/* calculates the entropy of the empirical measure */
double empirical_entropy() {
    int i;
    double e = 0.0;

    if( empirical.cached && empirical.stack_top < MAX_TOKEN_LINE_STACK) {
	for(i = 0; i < empirical.stack_top; i++) {
	    e += ((double)empirical.stack[i]->count) * log((double)empirical.stack[i]->count);
	}
    } else {
	for(i = 0; i < empirical.max_tokens; i++) {
	    if( FILLEDP(&empirical.hash[i]) ) {
		e += ((double)empirical.hash[i].count) * log((double)empirical.hash[i].count);
	    }
	}
    }
    e = e/empirical.full_token_count - log((double)empirical.full_token_count);


    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "empirical entropy = %g\n", -e);
    }

    return -e;
}

/***********************************************************
 * CATEGORY FUNCTIONS                                      *
 ***********************************************************/

/* initialize to zero. Filename specified elsewhere */
void init_category(Category *cat) {
    char *p;

    cat->full_token_count = 0;
    cat->unique_token_count = 0;
    cat->score = 0.0;
    if( (p = strrchr(cat->filename, '/')) ) {
	cat->filename = p + 1; /* only keep basename */
    }
}

/* turns purely random text into a category of its own */
void init_purely_random_text_category(Category *cat) {
    int i, j;
    double z = -log((double)ASIZE);
    
    for(i = 1; i < ASIZE; i++) {
	for(j = 0; j < ASIZE; j++) {
	    cat->dig[i][j] = z;
	}
    }
    z = -log((double)ASIZE - 1.0);
    for(j = 1; j < ASIZE; j++) {
	cat->dig[0][j] = z;
    }
    cat->dig[0][0] = log(0.0);
    cat->logZ = 0.0;
    cat->hash = NULL;
}

c_item *find_in_category(Category *cat, unsigned long int id) {
    c_item *i, *loop;

    if( cat->hash ) {
	/* start at id */
	i = loop = &cat->hash[id & (cat->max_tokens - 1)];

	while( FILLEDP(i) ) {
	    if( EQUALP(i->id,id) ) {
		return i; /* found id */
	    } else {
		i++; /* not found */
		i = (i >= &cat->hash[cat->max_tokens]) ? cat->hash : i; /* wrap around */
		if( i == loop ) {
		    return NULL; /* when hash table is full */
		}
	    }
	}
	return i;
    } else {
	return NULL;
    }
}

/* for each loaded category, this calculates the score */
void score_word(char *tok) {
    int i;
    int pp, pc;
    unsigned long int id;
    char *q;
    c_item *k;
    h_item *h;

    if( *tok ) {
	id = hash(tok, strlen(tok), 0);

	if( cat_count == 1 ) { /* calculate empirical distribution */
	    h = find_in_empirical(id);
	    if( h ) {
		h->count++;
		if( !FILLEDP(h) ) {
		    SET(h->id,id);
		    empirical.unique_token_count++;
		}
		if( empirical.cached && (empirical.stack_top < MAX_TOKEN_LINE_STACK)) {
		    empirical.stack[empirical.stack_top++] = h;
		}
	    }
	}
	empirical.full_token_count++; /* even if cat_count > 1 */

	/* now do scoring for all available categories */
	for(i = 0; i < cat_count; i++) {
	    k = find_in_category(&cat[i], id);
	    /* if token found, add its lambda weight */
	    if( k ) {
		cat[i].score += k->lam;
	    }
   
	    /* now compute the reference weight from digram model */
	    q = tok;
	    pp = 0;
	    while( *q ) {
		pc = *q - 96;
		cat[i].score += cat[i].dig[pp][pc];
		pp = pc;
		q++;
	    }
	    cat[i].score += cat[i].dig[pp][0]; /* trailing space */
	    cat[i].score -= cat[i].logZ;      /* don't forget the normalizing constant */
	}
    }
}

void line_score_single_category() {
    double score = 0.0;

    if( options & (1<<OPTION_POSTERIOR) ) {
      score = -cat[0].score/empirical.full_token_count;
      fprintf(stdout, "%8.2f %s", score, textbuf);
    }
    if( options & (1<<OPTION_SCORES) ) {
	score = -empirical_entropy() - cat[0].score/empirical.full_token_count;
	fprintf(stdout, "%8.2f %s", score, textbuf);
    }
    cat[0].score = 0.0;
    clear_empirical();
}

void line_score_multiple_categories() {
    int i, k;
    double c, cmin;

    cmin = -cat[0].score;
    k = 0;
    for(i = 1; i < cat_count; i++) {
	if(cmin > -cat[i].score) {
	    cmin = -cat[i].score;
	    k = i;
	}
    }

    if(options & (1<<OPTION_POSTERIOR) ) {
	/* compute probabilities given exclusive choices */
	c = 0.0;
	for(i = 0; i < cat_count; i++) {
	    c += exp((cat[i].score + cmin));
	}

	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %6.2f%% ", 
		    cat[i].filename, 100 * exp((cat[i].score + cmin))/c);
	}
	fprintf(stdout, "%s", textbuf);
    } else if( options & (1<<OPTION_SCORES) ) {
	/* display normalized divergence scores */
	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %6.2f ", 
		    cat[i].filename, -cat[i].score/empirical.full_token_count);
	}
	fprintf(stdout, "%s", textbuf);
    } else {
	/* prune the text which doesn't fit */
	for(i = 0; i < filter_count; i++) {
	    if( k == filter[i] ) {
		fprintf(stdout, "%s", textbuf);
		break;
	    }
	}
    }
    
    /* clean up for next line */
    for(i = 0; i < cat_count; i++) {
	cat[i].score = 0.0;
    }
    empirical.full_token_count = 0;
}
void score_single_category() {
    double score = 0.0;

    if( options & (1<<OPTION_POSTERIOR) ) {
      score = -cat[0].score/empirical.full_token_count;
      fprintf(stdout, "cross_entropy: %8.2f\n", score);
    }

    if( options & (1<<OPTION_SCORES) ) {
	score = -empirical_entropy() - cat[0].score/empirical.full_token_count;
	fprintf(stdout, "KL_score: %8.2f\n", score);
    }
    exit_code = 1;
}

void score_multiple_categories() {
    int i;
    double c, cmin;

    /* find the maximum likelihood category */
    cmin = -cat[0].score;
    exit_code = 0;
    for(i = 1; i < cat_count; i++) {
	if(cmin > -cat[i].score) {
	    cmin = -cat[i].score;
	    exit_code = i;
	}
    }

    /* here we compute probabilities given exclusive choices */
    if( options & (1<<OPTION_POSTERIOR) ) {

	c = 0.0;
	for(i = 0; i < cat_count; i++) {
	    c += exp((cat[i].score + cmin));
	}

	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %5.2f%% ", 
		    cat[i].filename, 100 * exp((cat[i].score + cmin))/c);
	}
	fprintf(stdout, "\n");
    }

    /* display normalized divergence scores */
    if( options & (1<<OPTION_SCORES) ) {
	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %5.2f ", 
		    cat[i].filename, -cat[i].score/empirical.full_token_count);
	}
	fprintf(stdout, "\n");
    }

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "%s\n", cat[exit_code].filename); 
    }

    exit_code++;
}

/***********************************************************
 * FILE MANAGEMENT FUNCTIONS                               *
 ***********************************************************/

/* used by both load and save functions */
#define MAGIC1 "# dbacl version "VERSION" category %s %s\n"
#define MAGIC2 "# model divergence: %lg logZ: %lg\n"
#define MAGIC3 "# hash_size: %hi full_token_count: %li unique_token_count: %li\n"

/* loads a category hash */
int load_category(Category *cat) {
    long int i, j;
    char buf[256];

    FILE *input;

    if( (input = fopen(cat->filename, "rb")) ) {
	fgets(buf, 256, input);
	if( strncmp(buf, MAGIC1, 16 + strlen(VERSION)) ) {
	    fprintf(stderr, "error: not a dbacl version "VERSION" category file [%s]\n", cat->filename);
	    return 0;
	} 
	fgets(buf, 256, input);
	sscanf(buf, MAGIC2, &cat->divergence, &cat->logZ);
	fgets(buf, 256, input);
	sscanf(buf, MAGIC3, &cat->max_token_bits, 
	       &cat->full_token_count, &cat->unique_token_count);
	cat->max_tokens = (1<<cat->max_token_bits);

	/* read character frequencies */
	fread(cat->dig, sizeof(float), ASIZE * ASIZE, input);

	/* allocate hash table */
	if( !(cat->hash = malloc(sizeof(c_item) * cat->max_tokens)) ) {
	    fprintf(stderr, "error: not enough memory for category %s\n", 
		    cat->filename);
	    return 0;
	}
	/* read in hash table */
	i = cat->max_tokens;
	j = 0;
	while( j < i ) {
	    j += fread(cat->hash + j, sizeof(c_item), i, input);
	}
	fclose(input);
    } else {
	fprintf(stderr, "error: cannot open file for reading %s\n", cat->filename);
	return 0;
    }

    return 1;
}


/* writes the learner to a file for easily readable category */
int save_learner() {
    long int i, j;
    FILE *output;
    c_item ci;

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "saving category to file %s\n", learner.filename);
    }
  
    if( (output = fopen(learner.filename, "wb")) ) {
	fprintf(output, MAGIC1, learner.filename, 
		(options & (1<<OPTION_REFMODEL)) ? "(ref)" : "");
	fprintf(output, MAGIC2, learner.divergence, learner.logZ);
	fprintf(output, MAGIC3, learner.max_token_bits, 
		learner.full_token_count, learner.unique_token_count);

	/* character frequencies */
	for(i = 0; i < ASIZE; i++) {
	    for(j = 0; j < ASIZE; j++) {
		while(fwrite(&learner.dig[i][j], sizeof(float), 1, output) < 1);
	    }
	}

	/* token/feature weights */
	for(i = 0; i < learner.max_tokens; i++) {
	    /* write each element so that it's easy to read back in a c_item array */
	    ci.id[0] = learner.hash[i].id[0];
	    ci.id[1] = learner.hash[i].id[1];      
	    ci.id[2] = learner.hash[i].id[2];
	    ci.lam = learner.hash[i].lam;
	    while(fwrite(&ci, sizeof(ci), 1, output) < 1); 
	}
	fclose(output);
    } else {
	fprintf(stderr, "error: cannot open file for writing %s\n", learner.filename);
	return 0;
    }

    return 1;
}


/***********************************************************
 * LEARNER FUNCTIONS                                       *
 ***********************************************************/

l_item *find_in_learner(unsigned long int id) {
    l_item *i, *loop;
    /* start at id */
    i = loop = &learner.hash[id & (learner.max_tokens - 1)];

    while( FILLEDP(i) ) {
	if( EQUALP(i->id,id) ) {
	    return i; /* found id */
	} else {
	    i++; /* not found */
	    i = (i >= &learner.hash[learner.max_tokens]) ? learner.hash : i; /* wrap around */
	    if( i == loop ) {
		return NULL; /* when hash table is full */
	    }
	}
    }

    /* empty slot, so not found */

    return i; 
}

/* places the token in the global hash and writes the
   token to a temporary file for later, then updates
   the digram frequencies */
void hash_word_and_learn(char *tok) {
    unsigned long int id;
    l_item *i;
    int p,q;

    if( *tok ) {
	id = hash(tok, strlen(tok), 0);
	i = find_in_learner(id);
	if( i ) {
	    if( FILLEDP(i) ) {
		i->count++; 
		learner.full_token_count++;
	    } else if( /* !FILLEDP(i) && */
		(100 * learner.unique_token_count < HASH_FULL * learner.max_tokens)) { 
		/* fill the hash and write to file */
		SET(i->id,id);
		learner.unique_token_count++;
		learner.full_token_count++;
		i->count++; 

		fputs(tok, learner.tmp);
		fputc(' ', learner.tmp);
	    }
	}

	/* now update digram frequency counts 
	   use zero to represent intertoken space */
	p = 0;
	/* finally update character frequencies */
	while( *tok ) {
	    q = *tok - 96;
	    learner.dig[p][q]++;
	    p = q;
	    tok++;
	}
	if( p ) { /* don't count digram 00 */ 
	    learner.dig[p][0]++;
	}
    }
}

/* initialize global learner object */
void init_learner() {
    int i, j;

    /* some constants */
    learner.max_tokens = default_max_tokens;
    learner.max_token_bits = default_max_token_bits;
    learner.full_token_count = 0;
    learner.unique_token_count = 0;
    learner.logZ = 0.0;

    /* init character frequencies - #26 is interword space */
    for(i = 0; i < ASIZE; i++) { 
	for(j = 0; j < ASIZE; j++) { 
	    learner.dig[i][j] = 0.0; 
	}
    }

    /* open temporary file for writing tokens */
    learner.tmp = tmpfile();

    /* allocate room for hash */
    learner.hash = calloc(learner.max_tokens, sizeof(l_item));
    if( !learner.hash ) {
	fprintf(stderr, "error: not enough memory? I couldn't allocate %li bytes\n", (sizeof(l_item) * learner.max_tokens));
	exit(0);
    }
}


/* calculates the normalizing constant */
double learner_Z() {
    long int i;
    long int c = 0;
    double t1 = 0.0;
    double t2 = 0.0;

    for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) ) {
	    t1 += exp(learner.hash[i].ref);
	    t2 += exp(learner.hash[i].lam + learner.hash[i].ref);
	    if( ++c >= learner.unique_token_count ) {
		c = 0;
		break;
	    }
	}
    }
    return (1.0 - t1) + t2;
}

/* calculates the divergence but needs normalizing constant */
double learner_divergence(double z) {
    long int i;
    long int c = 0;
    double t = 0.0;

    for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) ) {
	    t += learner.hash[i].lam * ((double)learner.hash[i].count/learner.full_token_count);
	    if( ++c >= learner.unique_token_count ) {
		c = 0;
		break;
	    }
	}
    }
    return -log(z) + t;
}

/* tolerance for the error in divergence. */
#define TOL 0.0005
/* minimizes the divergence by solving for lambda one 
   component at a time.  */
void minimize_learner_divergence() {
    long int i;
    long int c = 0;
    double d, dd, z, thresh, old_lam;

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "optimizing model divergence\n");
    }

    z = learner_Z();
    dd = learner_divergence(z);
    do {
	d = dd;     /* save old divergence */

	thresh = 0.0;
	/* change each lambda value one at a time,
	   since we only rarely recalculate Z,
	   we randomly decide to update a given lambda to
	   prevent large errors and periodic behaviour */
	for(i = 0; i < learner.max_tokens; i++) {
	    if( FILLEDP(&learner.hash[i]) ) {
		if( rand() > (RAND_MAX>>1)) {
		    old_lam = learner.hash[i].lam;
		    learner.hash[i].lam = log(z - exp(learner.hash[i].lam + learner.hash[i].ref)) *
			((double)learner.hash[i].count / ((double)(learner.full_token_count - learner.hash[i].count)))
			- learner.hash[i].ref;
		    thresh += fabs(old_lam - learner.hash[i].lam) * learner.hash[i].ref;
		}
		if( thresh > learner.full_token_count ) {
		    z = learner_Z();
		    thresh = 0.0;
		}
		if( ++c >= learner.unique_token_count ) {
		    c = 0;
		    break;
		}
	    }
	}
	/* update values */
	z = learner_Z();
	dd = learner_divergence(z);

	if( options & (1<<OPTION_VERBOSE) ) {
	    fprintf(stdout, "divergence change %f --> %f\n",d, dd);
	}

    } while( fabs(d - dd) > TOL );

    /* now save values */
    learner.logZ = log(z);
    learner.divergence = dd;
}

/* initializes to the Laplacian parameters */ 
void init_dirichlet() {
    int i;

    dirichlet.alpha = 1.0;
    for(i = 0; i < ASIZE; i++) {
	dirichlet.u[i] = 1.0/((double)ASIZE);
    }
}

/* calculates the most probable Dirichlet parameters 
   given digram counts. Method described in MacKay and Peto (1995)
   Since we don't count the 00 digram, including the prior will 
   overestimate the 00 transition probability. However, for this 
   transition as well as the others, the Dirichlet will have practically
   no effect when there's enough data in the corpus.
*/
void optimize_dirichlet() {
    int i, j;
    long int k;
    float G[ASIZE];
    float H[ASIZE];
    float V[ASIZE];
    float F[ASIZE];
    double prev_alpha;
    double K, t;

    /* precompute useful constants */
    for(j = 0; j < ASIZE; j++) {
	G[j] = 0.0;
	H[j] = 0.0;
	V[j] = 0.0;
	F[j] = 0.0;
	for(i = 0; i < ASIZE; i++) {
	    for(k = 1; k < learner.dig[i][j]; k++) {
		G[j] += 1.0/k;
		H[j] += 1.0/((double)k*k);
	    }
	    if( learner.dig[i][j] > 0 ) { 
		V[j]++; 
		F[j] += learner.dig[i][j];
	    }
	}
	if( options & (1<<OPTION_DEBUG) ) { 
	    fprintf(stdout, "symbol %c has G = %g H = %g V = %g F = %g\n", j + 96, G[j], H[j], V[j], F[j]);
	}
    }

    /* now optimize the dirichlet parameters */
    do {
	prev_alpha = dirichlet.alpha;
	K = 0.0;
	for(i = 0; i < ASIZE; i++) {
	    K += log((F[i] + prev_alpha)/prev_alpha) +
		(0.5) * F[i]/(prev_alpha * (F[i] + prev_alpha));
	}
	dirichlet.alpha = 0.0;
	for(j = 0; j < ASIZE; j++) {
	    t = K - G[j];
	    dirichlet.u[j] = 2 * V[j] /(t + sqrt(t * t + 4 * H[j] * V[j]));
	    dirichlet.alpha += dirichlet.u[j];
	    //fprintf(stderr, "D = %g m[%i] = %g ", t * t - 4 * H[j] * V[j], j, dirichlet.m[j]);
	    //fprintf(stderr, "info: dirichlet alpha %g --> %g\n", prev_alpha, dirichlet.alpha);
	}
	if( options & (1<<OPTION_VERBOSE) ) {
	    fprintf(stdout, "dirichlet alpha %g --> %g\n", prev_alpha, dirichlet.alpha);
	}
    } while( fabs(prev_alpha - dirichlet.alpha) > TOL );

    if( options & (1<<OPTION_DEBUG) ) {
	for(j = 0; j < ASIZE; j++) {
	    fprintf(stdout, "dirichlet u[%c] =  %g\n", j + 96, dirichlet.u[j]);
	}
    }
}

/* computes the _logarithm_ of digram transition probabilities */
void compute_digram_probabilities() {
    int i,j;
    double t;
    
    for(i = 0; i < ASIZE; i++) {
	t = 0.0;
	for(j = 0; j < ASIZE; j++) {
	    t += learner.dig[i][j];
	}
	for(j = 0; j < ASIZE; j++) {
	    learner.dig[i][j] = 
		log(((double)learner.dig[i][j] + dirichlet.u[j]) / 
		    (t + dirichlet.alpha));
	    if( options & (1<<OPTION_DEBUG) ) {
		fprintf(stdout,"learner.dig[%c][%c] = %f\n", i + 96, j + 96, learner.dig[i][j]);
	    }
	}
    }
}

void optimize_learner_and_save() {
    long int i;
    double t;
    unsigned long int id;
    l_item *k;

    char buf[BUFLEN];
    char tok[MAX_TOKEN_LEN];
    
    char *p;
    char *q = tok;
    int pp, pc;

    if(100 * learner.unique_token_count >= HASH_FULL * learner.max_tokens) { 
	fprintf(stderr, "warning: table full, some tokens ignored - try with option -h %i\n",
		learner.max_token_bits + 1);
    }

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "picked up %li (%li distinct) tokens\n", 
		learner.full_token_count, learner.unique_token_count);
	fprintf(stdout, "calculating reference word weights\n");
    }

    /* find best Dirichlet parameters */
    init_dirichlet();
    optimize_dirichlet();

    /* estimate the conditional digram probabilities */
    compute_digram_probabilities();

    /* now we calculate the logarithmic word weight
       from the digram model, for each token in the hash */
    rewind(learner.tmp);
    t = 0.0;
    pp = 0;
    while( !feof(learner.tmp) && fgets(buf, BUFLEN, learner.tmp) ) {
	p = buf;
	while( *p ) {
	    if( *p != ' ') {
		pc = *p - 96;
		t += learner.dig[pp][pc];
		*q++ = *p; /* copy into tok */
	    } else { /* interword space */ 
		pc = 0;
		t += learner.dig[pp][pc];
		*q = 0; /* append NUL to tok */

		/* now write weight in hash */
		id = hash(tok, strlen(tok), 0);
		k = find_in_learner(id); /* guaranteed to be found */
		k->ref = t;
	
		if( options & (1<<OPTION_DEBUG) ) {
		    fprintf(stdout, "%s ref_score %g (random ref_score %g)\n", tok, t, - log((double)ASIZE - 1.0) - strlen(tok) * log((double)ASIZE));
		}
		/* prepare for next token */
		t = 0.0;
		q = tok; /* reset q */
	    }
	    pp = pc;
	    p++;
	}
    }
    /* finish last word */
    t += learner.dig[pp][0];
    *q = 0;
    id = hash(tok, strlen(tok), 0);
    k = find_in_learner(id); /* guaranteed to be found */
    k->ref = t;

    /* done */
    fclose(learner.tmp);

    if( options & (1<<OPTION_VERBOSE) ) {
	t = 0.0;
	for(i=0; i < learner.max_tokens; i++) {
	    if( FILLEDP(&learner.hash[i]) ) {
		t += exp(learner.hash[i].ref);
	    }
	}
	fprintf(stdout, "tokens account for %g of reference probability mass\n", t);
    }

    /* now we have the reference measure and token frequencies
       so we can optimize divergence and calculate normalizing constant */

    if( !(options & (1<<OPTION_REFMODEL))) {
	minimize_learner_divergence();
    }

    /* now save the model to a file */
    save_learner();
}


/***********************************************************
 * MAIN FUNCTIONS                                          *
 ***********************************************************/

/* reads a BSD mbox file (email format) and processes only the
   flat ascii text within it, skipping mail headers and
   assorted MIME types - the MIME parsing is quick and dirty */
void process_mbox_file(FILE *input, void (*word_fun)(), 
		       void (*line_fun)()) {
    char tok[MAX_TOKEN_LEN];

    char *p;
    char *q = tok;
    char *s;
    int k, l;

    mbox_state state = UNDEF;
    int process_line = 0;
    int prev_line_empty = 1;
    int mime_type = 0;
    int checked_content_type = 0;
    xml_state xstate;
    xstate = TEXT;
    /* apply wordfunc on each alpha token */
    while( !feof(input) ) {
	/* read in a full line, allocating memory as necessary */
	s = textbuf;
	l = textbuf_len;
	k = 1;
	while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
	    textbuf = realloc(textbuf, 2 * textbuf_len);
	    s = textbuf + textbuf_len - (k++);
	    l = textbuf_len;
	    textbuf_len *= 2;
	}
	/* below we decide if we want to process the line */
	process_line = 0; /* by default we skip the line */
	switch(state) {
	case UNDEF:
	    if( prev_line_empty && 
		(!strncmp(textbuf, "From ", 5) || 
		 (!strncmp(textbuf, "--", 2) && (textbuf[2] != ' '))) ) {
		/* if it starts with From_ or else it looks like --xxxxx, 
		   it signals a new header */
		state = HEADER;
		checked_content_type = 0;
	    }
	    break;
	case HEADER:
	    if( *textbuf == '\n' ) {
		state = BODY;
	    } else if( !strncmp(textbuf, "Content-Type:", 13) ) {
		checked_content_type = 1;
		if( strstr(textbuf + 13, "text/") || strstr(textbuf + 13, "TEXT/") ) {
		    mime_type = 1; /* good */
		} else {
		    mime_type = 0; /* bad */
		}
	    } else if( !strncmp(textbuf, "Subject:", 8) ||
		       !(strncmp(textbuf, "From:", 5)) ) {
	      /* process subject and from line like body */
	      process_line = 1;
	    }
	    break;
	case BODY:
	    if( prev_line_empty && 
		(!strncmp(textbuf, "From ", 5) || 
		 (!strncmp(textbuf, "--", 2) && !strchr(textbuf + 2, ' ')))) {
		/* if it starts with From_ or else it looks like --xxxxx, 
		   it signals a new header */
		state = HEADER;
		checked_content_type = 0;
	    } else if( !(checked_content_type && !mime_type) ) { 
		/* unless we explicitly are told it's not text, we process it */
		process_line = 1;
	    }
	    break;
	}
	prev_line_empty = (*textbuf == '\n') ? 1 : 0; /* for next time */
	/* now step through the line one char at a time */
	if( process_line ) {
	    p = textbuf;
	    while( *p ) {
	      switch(xstate) {
	      case TEXT:
		if( isalpha(*p) && (q < tok + MAX_TOKEN_LEN)) {
		    *q++ = tolower(*p);
		} else {
		    *q = 0;
		    q = tok;
		    (*word_fun)(q);
		    if( (options & (1<<OPTION_XML)) && 
			(*p == '<') &&
			!(state == HEADER) ) {
		      xstate = TAG;
		    }
		}
		break;
	      case TAG:
		if( *p == '>' ) {
		  xstate = TEXT;
		}
		break;
	      }
	      p++;
	    }
	    if( line_fun ) { (*line_fun)(); }
	} else {
	  xstate = TEXT; /* reset XML tag state */
	}
    }
}

/* reads a flat ascii text file as input and operates on
   each word found by calling wordfunc().
   If required, makes a half hearted attempt to skip XML tags. Since
   we're interested in statistics of alphabetic tokens only, mistakes
   in XML parsing aren't very critical */
void process_text_file(FILE *input, void (*word_fun)(), 
		       void (*line_fun)()) {
    char tok[MAX_TOKEN_LEN];

    char *p;
    char *q = tok;
    char *s;
    int k, l;
    xml_state state;
    
    state = TEXT;
    /* apply wordfunc on each alpha token */
    while( !feof(input) ) {
	/* read in a full line, allocating memory as necessary */
	s = textbuf;
	l = textbuf_len;
	k = 1;
	while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
	    textbuf = realloc(textbuf, 2 * textbuf_len);
	    s = textbuf + textbuf_len - (k++);
	    l = textbuf_len;
	    textbuf_len *= 2;
	}
	/* now step through the line one char at a time */
	p = textbuf;
	while( *p ) {
	  switch(state) {
	  case TEXT:
	    if( isalpha(*p) && (q < tok + MAX_TOKEN_LEN)) {
		*q++ = tolower(*p);
	    } else {
		*q = 0;
		q = tok;
		(*word_fun)(q);
		if( (options & (1<<OPTION_XML)) && (*p == '<') ) {
		  state = TAG;
		}
	    }
	    break;
	  case TAG:
	    if( *p == '>' ) {
	      state = TEXT;
	    }
	    break;
	  }
	  p++;
	}
	if( line_fun ) { (*line_fun)(); }
    }
}

char *sanitize_path(char *in) {
  char *q;
  char *path;
  int l;

  if( (*in != '/') && (path = getenv("DBACL_PATH")) ) {
    l = strlen(path);
    q = malloc(l + strlen(in) + 3);
    strcpy(q, path);
    if( q[l - 1] != '/' ) {
      q[l] = '/'; 
      q[l + 1] = 0;
    }
    strcat(q, in);
    return q;
  } else {
    return in;
  }
}

int main(int argc, char **argv) {

    int k;
    FILE *input;
    char op;

    void (*preprocess_fun)() = NULL;
    void (*word_fun)() = NULL;
    void (*line_fun)() = NULL;
    void (*postprocess_fun)() = NULL;

    /* parse the options */
    while( (op = getopt(argc, argv, "DRVvNnrh:c:l:T:f:")) > -1 ) {

      switch(op) {
      case 'D':
	options |= (1<<OPTION_DEBUG);
	break;
      case 'N':
	options |= (1<<OPTION_POSTERIOR);
	break;
      case 'R':
	if( cat_count >= MAX_CAT ) {
	  fprintf(stderr, "warning: maximum reached, random text category omitted\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, "error: cannot use options -l and -R together\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_CLASSIFY);

	  cat[cat_count].filename = "random";

	  init_category(&cat[cat_count]);
	  init_purely_random_text_category(&cat[cat_count]);

	  cat_count++;
	}
	break;
      case 'T':
	if( !strncasecmp(optarg, "email", 5) ) {
	  options |= (1<<OPTION_MBOX_FORMAT);
	  options |= (1<<OPTION_XML); /* filter out HTML from messages */
	} else if( !strncasecmp(optarg, "xml", 3) ) {
	  options |= (1<<OPTION_XML);
	} else { /* default */
	  options |= (1<<OPTION_TEXT_FORMAT);
	}
	break;
      case 'V':
	fprintf(stdout, "dbacl version %s\n", VERSION);
	exit(1);
	break;
      case 'h': /* select memory size in megabytes */
	default_max_token_bits = strtod(optarg, NULL);
	if( default_max_token_bits > MAX_TOKEN_BITS ) {
	  fprintf(stderr, "warning: maximum hash size will be 2^%d\n", MAX_TOKEN_BITS);
	  default_max_token_bits = MAX_TOKEN_BITS;
	}
	default_max_tokens = (1<<default_max_token_bits);
	break;
      case 'n':
	options |= (1<<OPTION_SCORES);
	break;
      case 'c':
	if( cat_count >= MAX_CAT ) {
	  fprintf(stderr, "warning: maximum reached, category ignored\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, "error: cannot use options -l and -c together\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_CLASSIFY);

	  cat[cat_count].filename = sanitize_path(optarg);

	  if( !*optarg ) {
	    fprintf(stderr, "error: category needs a name\n");
	    usage(argv);
	    exit(0);
	  }
	  if( load_category(&cat[cat_count]) ) {
	    init_category(&cat[cat_count]);
	  } else {
	    fprintf(stderr, "error: could not load category %s\n", cat[cat_count].filename);
	    exit(0);
	  }
	  cat_count++;
	}
	break;
      case 'r':
	options |= (1<<OPTION_REFMODEL);
	break;
      case 'f':
	if( filter_count >= MAX_CAT ) {
	  fprintf(stderr, "warning: maximum reached, filter ignored\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, "error: cannot use options -l and -f together\n");
	  usage(argv);
	  exit(0);
	} else if( !*optarg ) {
	  fprintf(stderr, "error: filter must be category name or number\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_FILTER);
	  filter[filter_count] = -1;

	  /* see if it's a known category */
	  for(k = 0; k < cat_count; k++) {
	    if( !strcmp(cat[k].filename, optarg) ) {
	      filter[filter_count] = k;
	      break;
	    }
	  }
	  /* if not recognized, see if it's a number */
	  if( filter[filter_count] < 0 ) {
	    filter[filter_count] = strtod(optarg, NULL) - 1;
	  }
	  if( filter[filter_count] < 0 ) { /* still not recognized */
	    fprintf(stderr, "error: unrecognized category in -f option [%s]\n", optarg);
	    usage(argv);
	    exit(0);
	  }
	  filter_count++;
	}
	break;
      case 'v':
	options |= (1<<OPTION_VERBOSE);
	break;
      case 'l': 
	if( options & (1<<OPTION_CLASSIFY) ) {
	  fprintf(stderr, "error: cannot use options -l and -c together\n");
	  usage(argv);
	  exit(0);
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, "error: option -l can only occur once\n");
	  exit(0);
	} else {
	  options |= (1<<OPTION_LEARN);
	  learner.filename = sanitize_path(optarg);
	  if( !*learner.filename ) {
	    fprintf(stderr, "error: category needs a name\n");
	    usage(argv);
	    exit(0);
	  }
	}
	break;
      default:
	break;
      }
    }
    
    /* consistency check */
    if( ((options>>OPTION_CLASSIFY) & 1) + 
	((options>>OPTION_LEARN) & 1) != 1 ) {
      fprintf(stderr, "error: please use either -c or -l option\n");
      usage(argv);
      exit(0);
    }

    if( ((options>>OPTION_TEXT_FORMAT) & 1) + 
	((options>>OPTION_MBOX_FORMAT) & 1) > 1 ) {
      fprintf(stderr, "error: please use one of either -T text or -T email options\n");
      usage(argv);
      exit(0);
    }

    /* set up callbacks */
    if( options & (1<<OPTION_CLASSIFY) ) {
      if( cat_count == 1 ) {
	/* single category */

	init_empirical(); /* sets cached to zero */
	    
	preprocess_fun = NULL;
	word_fun = score_word;
	if( options & (1<<OPTION_FILTER) ) {
	  empirical.cached = 1;
	  line_fun = line_score_single_category;
	  postprocess_fun = NULL;
	} else {
	  line_fun = NULL;
	  postprocess_fun = score_single_category;
	}

      } else {
	/* multiple categories */

	empirical.full_token_count = 0; /* we don't any other part of empirical */

	preprocess_fun = NULL;
	word_fun = score_word;
	if( options & (1<<OPTION_FILTER) ) {
	  empirical.cached = 1;
	  line_fun = line_score_multiple_categories;
	  postprocess_fun = NULL;
	} else {
	  line_fun = NULL;
	  postprocess_fun = score_multiple_categories;
	}

      }
    } else if( options & (1<<OPTION_LEARN) ) {
      /* category learning */

      preprocess_fun = init_learner;
      word_fun = hash_word_and_learn;
      line_fun = NULL;
      postprocess_fun = optimize_learner_and_save;

    } else { /* something wrong ? */
      usage(argv);
      exit(0);
    }

    if( preprocess_fun ) { (*preprocess_fun)(); }

    textbuf_len = BUFLEN;
    textbuf = malloc(textbuf_len);
    /* now process each file on the command line,
       or if none provided read stdin */
    while( (optind > -1) && *(argv + optind) ) {
      /* if it's a filename, process it */
      if( (input = fopen(argv[optind], "r")) ) {

	options |= (1<<INPUT_FROM_CMDLINE);

	if( (options & (1<<OPTION_VERBOSE)) && !(options & (1<<OPTION_CLASSIFY))) {
	  fprintf(stdout, "processing file %s\n", argv[optind]);
	}

	if( options & (1<<OPTION_MBOX_FORMAT) ) {
	  process_mbox_file(input, word_fun, line_fun);
	} else { /* default to text */
	  process_text_file(input, word_fun, line_fun);
	}

	fclose(input);

      } else { /* unrecognized file name */

	fprintf(stderr, "error: couldn't open %s\n", argv[optind]);
	usage(argv);
	exit(0);

      }
      optind++;
    }
    /* in case no files were specified, get input from stdin */
    if( !(options & (1<<INPUT_FROM_CMDLINE)) ) {

      if( (options & (1<<OPTION_VERBOSE)) && !(options & (1<<OPTION_CLASSIFY)) ) {
	fprintf(stdout, "taking input from stdin\n");
      }

      if( options & (1<<OPTION_MBOX_FORMAT) ) {
	process_mbox_file(stdin, word_fun, line_fun);
      } else { /* default to text */
	process_text_file(stdin, word_fun, line_fun);
      }

    }
  
    if( postprocess_fun ) { (*postprocess_fun)(); }

    /* free some global resources */
    free(textbuf);

    exit(exit_code);
}
